# Generated by Django 4.1.2 on 2022-11-09 23:37

from django.db import migrations
from django.core.management import call_command

fixture = 'users'


def load_fixture(apps, schema_editor):
    call_command('loaddata', fixture, app_label='users')


def unload_fixture(apps, schema_editor):
    User = apps.get_model("user", "User")
    User.objects.all().delete()


class Migration(migrations.Migration):

    dependencies = [
        ("users", "0001_initial"),
    ]

    operations = [
        migrations.RunPython(load_fixture, reverse_code=unload_fixture),
    ]
